{
  "cells": [
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n# Heagy et al., 2017 Casing Example\n\nHere, we use a primary-secondary approach to compute the sensitivity of an EM\nsurvey with respect to a parametric model of a block in a layered space.\n\nIf you run this example with :code:`reRun=False`, stored results for the\nsensitivity will be downloaded and plotted. Otherwise, if `reRun=True`, the\nexample will be re-run. Note that you will need modest computational resources\nto re-compute the sensitivity (it is a 3D EM problem!).\n\nThere are 6 plots that are produced. Figure 1 shows the primary current density\n(the current density due to the source, casing and layered background), Figure\n2 shows the secondary source - the source current density due to the 3D\nconductivity structure present only in the secondary problem. Figure 3 shows\nthe predicted data. Figures 4-6 show the sensitivity with respect to each of\nthe 9 model parameters (the conductivity of the background, layer and block,\nthe thickness of the layer and the x and y extents of the block).\n\nThis example is used in the paper\n\n    Lindsey J. Heagy, Rowan Cockett, Seogi Kang, Gudni K. Rosenkjaer, Douglas\n    W. Oldenburg, A framework for simulation and inversion in electromagnetics,\n    Computers & Geosciences, Volume 107, 2017, Pages 1-19, ISSN 0098-3004,\n    http://dx.doi.org/10.1016/j.cageo.2017.06.018.\n\nThis example is available on figshare:\nhttps://doi.org/10.6084/m9.figshare.5036123\n\nThis example was updated for SimPEG 0.14.0 on January 31st, 2020 by Joseph Capriotti\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import discretize\nfrom simpeg import utils, maps, tests\nfrom simpeg.electromagnetics import frequency_domain as FDEM, mu_0\nfrom simpeg.utils.io_utils import download\n\n# try:\n#     from pymatsolver import MumpsSolver as Solver\n#     print('using MumpsSolver')\n# except ImportError:\ntry:\n    from pymatsolver import Pardiso as Solver\nexcept ImportError:\n    from simpeg import SolverLU as Solver\n\nimport numpy as np\nimport scipy.sparse as sp\nimport time\nimport os\n\nimport matplotlib.pyplot as plt\nfrom matplotlib.colors import LogNorm\nfrom matplotlib import rcParams\nimport h5py\n\nnp.random.seed(42)\nfontsize = 12\nrcParams[\"font.size\"] = fontsize\n\n\nclass PrimSecCasingExample(object):\n    NAME = \"PrimSec_5e6Casing_50Mu_05Hz_LargeCondBody\"\n\n    # -------------- SETUP MODEL PARAMS ---------------------------- #\n\n    sigmaair = 1e-8  # air\n    sigmaback = 1e-2  # background\n\n    sigmacasing = 5.5e6  # casing\n    sigmainside = 1  # inside the casing\n    mucasing = 50  # casing permeability\n\n    casing_l = 1000  # length of the casing\n    casing_d = 10e-2  # 10cm diameter casing\n    casing_t = 1e-2  # 1cm thickness\n\n    # layer\n    sigmalayer = 1.0 / 10.0\n    layer_z = np.r_[-1000.0, -900.0]\n\n    # 3D body\n    sigmablock = 2.0\n    block_x = np.r_[75.0, 475.0]\n    block_y = np.r_[-125, 125.0]\n    block_z = layer_z\n\n    # Survey Params\n    freqs = np.r_[0.5]  # frequencies 0.5\n    dsz = -950.0  # down-hole z source location\n\n    src_a = np.r_[0.0, 0.0, dsz]\n    src_b = np.r_[1e4, 0.0, 0.0]  # return electrode is a ring, 200\n\n    def __init__(self):\n        # translate casing parameters to radial distances to outer casing wall,\n        # inner casing wall, and casing length to z locations\n        self.casing_r = self.casing_d / 2.0\n        self.casing_a = self.casing_r - self.casing_t / 2.0  # inner radius\n        self.casing_b = self.casing_r + self.casing_t / 2.0  # outer radius\n        self.casing_z = np.r_[-self.casing_l, 0.0]\n\n        # Display skin depth so we can ensure our mesh goes further.\n        print(\n            \"\\nSkin Depth: {}\".format(\n                [(500.0 / np.sqrt(self.sigmaback * _)) for _ in self.freqs]\n            )\n        )\n\n    # -------------- Model --------------------------------- #\n    @property\n    def mtrue(self):\n        # This is the model we are using to compute the sensitivity. Each of\n        # these parameters would be considered unknown in an inversion. This\n        # model is of a parametrized block in a layer\n        #\n        #   +--------------------------------------+\n        #   |                                      |\n        #   |                                      |\n        #   |            background                |\n        #   |                                      |\n        #   |                                      |\n        #   +--------------------+-------+---------+\n        #   |                    |       |         |\n        #   |      layer         | block |         |\n        #   |                    |       |         |\n        #   +--------------------+-------+---------+\n        #   |                                      |\n        #   |                                      |\n        #   |            background                |\n        #   |                                      |\n        #   |                                      |\n        #   +--------------------------------------+\n\n        return np.hstack(\n            np.r_[\n                np.log(self.sigmaback),  # value in background\n                np.log(self.sigmalayer),  # value in the layer\n                np.log(self.sigmablock),  # value in the block\n                self.layer_z.mean(),  # layer center\n                self.layer_z[1] - self.layer_z[0],  # layer thickness\n                self.block_x.mean(),  # block x_0\n                self.block_y.mean(),  # block y_0\n                self.block_x[1] - self.block_x[0],  # block dx\n                self.block_y[1] - self.block_y[0],  # block dy\n            ]\n        )\n\n    # ----------------------------------------------------------------- #\n    # -------------- PRIMARY PROBLEM SETUP ---------------------------- #\n    # ----------------------------------------------------------------- #\n\n    @property\n    def meshp(self):\n        if getattr(self, \"_meshp\", None) is None:\n            # -------------- Mesh Parameters ------------------ #\n            # x-direction\n            csx1, csx2 = 2.5e-3, 25.0  # fine cells near well bore\n            pfx1, pfx2 = 1.3, 1.4  # padding factors: fine -> uniform\n            ncx1 = np.ceil(self.casing_b / csx1 + 2)  # number of fine cells\n            # (past casing wall)\n            dx2 = 1000.0  # uniform mesh out to here\n            npadx2 = 21  # padding out to infinity\n\n            # z-direction\n            csz = 0.05  # finest z-cells\n            nza = 10  # number of fine cells above air-earth interface\n            pfz = pfx2  # padding factor in z-direction\n\n            # ------------- Assemble the Cyl Mesh ------------- #\n            # pad nicely to second cell size\n            npadx1 = np.floor(np.log(csx2 / csx1) / np.log(pfx1))\n            hx1a = utils.unpack_widths([(csx1, ncx1)])\n            hx1b = utils.unpack_widths([(csx1, npadx1, pfx1)])\n            dx1 = sum(hx1a) + sum(hx1b)\n            dx1 = np.floor(dx1 / csx2)\n            hx1b *= (dx1 * csx2 - sum(hx1a)) / sum(hx1b)\n\n            # second chunk of mesh\n            ncx2 = np.ceil((dx2 - dx1) / csx2)\n            hx2a = utils.unpack_widths([(csx2, ncx2)])\n            hx2b = utils.unpack_widths([(csx2, npadx2, pfx2)])\n            hx = np.hstack([hx1a, hx1b, hx2a, hx2b])\n\n            # cell size, number of core cells, number of padding cells in the\n            # x-direction\n            ncz = int(np.ceil(np.diff(self.casing_z)[0] / csz)) + 10\n            npadzu, npadzd = 43, 43\n\n            # vector of cell widths in the z-direction\n            hz = utils.unpack_widths(\n                [(csz, npadzd, -pfz), (csz, ncz), (csz, npadzu, pfz)]\n            )\n\n            # primary mesh\n            self._meshp = discretize.CylindricalMesh(\n                [hx, 1.0, hz], [0.0, 0.0, -np.sum(hz[: npadzu + ncz - nza])]\n            )\n\n            print(\n                \"Cyl Mesh Extent xmax: {},: zmin: {}, zmax: {}\".format(\n                    self._meshp.cell_centers_x.max(),\n                    self._meshp.cell_centers_z.min(),\n                    self._meshp.cell_centers_z.max(),\n                )\n            )\n\n        return self._meshp\n\n    @property\n    def indActivePrimary(self):\n        return self.meshp.gridCC[:, 2] <= 0.0  # air cells\n\n    @property\n    def projectionMapPrimary(self):\n        return maps.Projection(nP=9, index=np.r_[0, 1, 3, 4])\n\n    @property\n    def primaryMapping(self):\n        # Setup Pimary Maps:\n        # we want to simulate on a physical property model that\n        # consists of casing in a layered background. Air cells are included.\n        # Our \"model\", that we are considering when computing the sensitivity,\n        # consists of the layered background and block, so the casing and air\n        # cells are inactive parts of the model and need to be appropriately\n        # injected during the construction of the primary model\n\n        if getattr(self, \"_primaryMapping\", None) is None:\n            print(\"Building primary mapping\")\n\n            # inject parameters we want to invert for into the full casing\n            # model\n            valInactive = np.r_[\n                np.log(self.sigmacasing),  # log conductivity of the casing\n                np.log(self.sigmainside),  # log conductivity fluid inside\n                # casing\n                self.casing_r,  # radius of the casing (to its center)\n                self.casing_t,  # casing thickness\n                self.casing_z[0],  # bottom of casing (at depth)\n                self.casing_z[1],  # top of casing (at surface)\n            ]\n\n            # inject casing parameters so they are included in the construction\n            # of the layered background + casing\n            injectCasingParams = maps.InjectActiveCells(\n                None, indActive=np.r_[0, 1, 4, 5], valInactive=valInactive, nC=10\n            )\n\n            # maps a list of casing parameters to the cyl mesh (below the\n            # subsurface)\n            paramMapPrimary = maps.ParametricCasingAndLayer(\n                self.meshp, indActive=self.indActivePrimary, slopeFact=1e4\n            )\n\n            # inject air cells\n            injActMapPrimary = maps.InjectActiveCells(\n                self.meshp, self.indActivePrimary, np.log(self.sigmaair)\n            )\n\n            # map from log conductivity to conductivity\n            expMapPrimary = maps.ExpMap(self.meshp)\n\n            # assemble the primary mapping\n            primaryMapping = (\n                expMapPrimary\n                * injActMapPrimary  # log(sigma) --> sigma\n                * paramMapPrimary  # log(sigma) below surface --> include air\n                * injectCasingParams  # parametric --> casing + layered earth  # parametric layered earth --> parametric\n                *\n                # layered earth + casing\n                self.projectionMapPrimary  # grab relevant parameters from full\n                # model (eg. ignore block)\n            )\n\n            self._paramMapPrimary = paramMapPrimary\n            self._primaryMapping = primaryMapping\n\n            print(\"... done building primary mapping\")\n\n        return self._primaryMapping\n\n    @property\n    def muModel(self):\n        # Mu Model\n        # here, we want to consider variable magnetic permeability in the\n        # simulation. The only permeable item in the domain is the casing.\n        if getattr(self, \"_muModel\", None) is None:\n            if getattr(self, \"_paramMapPrimary\", None) is None:\n                self.primaryMapping\n\n            muMap = (\n                maps.InjectActiveCells(self.meshp, self.indActivePrimary, mu_0)\n                * self._paramMapPrimary\n            )\n\n            muModel = muMap * np.hstack(\n                np.r_[\n                    mu_0,  # val Background\n                    mu_0,  # val Layer\n                    mu_0 * self.mucasing,  # val Casing\n                    mu_0,  # val inside Casing\n                    self.layer_z.mean(),  # layer center\n                    self.layer_z[1] - self.layer_z[0],  # layer thickness\n                    self.casing_r,  # casing radius\n                    self.casing_t,  # casing thickness\n                    self.casing_z[0],  # casing bottom\n                    self.casing_z[1],  # casing top\n                ]\n            )\n            self._muModel = muModel\n        return self._muModel\n\n    @property\n    def primaryProblem(self):\n        if getattr(self, \"_primaryProblem\", None) is None:\n            # define a custom prop map to include variable mu that we are not\n            # inverting for - This will change when we improve the propmap!\n            print(\"Getting Primary Problem\")\n\n            # class CasingEMPropMap(maps.PropMap):\n\n            #     sigma = maps.Property(\n            #                 \"Electrical Conductivity\", defaultInvProp=True,\n            #                 propertyLink=('rho', maps.ReciprocalMap)\n            #     )\n            #     mu = maps.Property(\n            #             \"Inverse Magnetic Permeability\",\n            #             defaultVal=self.muModel,\n            #             propertyLink=('mui', maps.ReciprocalMap)\n            #     )\n            #     rho = maps.Property(\n            #             \"Electrical Resistivity\",\n            #             propertyLink=('sigma', maps.ReciprocalMap)\n            #     )\n            #     mui = maps.Property(\n            #             \"Inverse Magnetic Permeability\",\n            #             defaultVal=1./self.muModel,\n            #             propertyLink=('mu', maps.ReciprocalMap)\n            #     )\n\n            # # set the problem's propmap\n            # FDEM.Simulation3DMagneticField.PropMap = CasingEMPropMap\n\n            # use H-J formulation for source with vertical current density and\n            # cylindrical symmetry (h faster on cyl --> less edges than faces)\n            primaryProblem = FDEM.Simulation3DMagneticField(\n                self.meshp, sigmaMap=self.primaryMapping\n            )\n            primaryProblem.mu = self.muModel\n\n            primaryProblem.solver = Solver\n            self._primaryProblem = primaryProblem\n\n            print(\"... done building primary problem\")\n\n        return self._primaryProblem\n\n    @property\n    def primarySurvey(self):\n        if getattr(self, \"_primarySurvey\", None) is None:\n            print(\"Setting up primary survey\")\n\n            def setupPrimarySource(plotIt=False):\n                # Construct a downhole source that is coupled to the casing\n                meshp = self.meshp\n                src_a = self.src_a\n                src_b = self.src_b\n                casing_a = self.casing_a\n\n                # downhole source\n                dg_x = np.zeros(meshp.vnF[0], dtype=complex)\n                dg_y = np.zeros(meshp.vnF[1], dtype=complex)\n                dg_z = np.zeros(meshp.vnF[2], dtype=complex)\n\n                # vertically directed wire in borehole\n                # go through the center of the well\n                dgv_indx = meshp.gridFz[:, 0] < meshp.h[0].min()\n                dgv_indz = (meshp.gridFz[:, 2] >= src_a[2]) & (\n                    meshp.gridFz[:, 2] <= src_b[2]\n                )\n                dgv_ind = dgv_indx & dgv_indz\n                dg_z[dgv_ind] = -1.0\n\n                # couple to the casing downhole - top part\n                dgh_indx = meshp.gridFx[:, 0] <= casing_a + meshp.h[0].min() * 2\n\n                # couple to the casing downhole - bottom part\n                dgh_indz2 = (meshp.gridFx[:, 2] <= src_a[2]) & (\n                    meshp.gridFx[:, 2] > src_a[2] - meshp.h[2].min()\n                )\n                dgh_ind2 = dgh_indx & dgh_indz2\n                dg_x[dgh_ind2] = 1.0\n\n                # horizontally directed wire\n                sgh_indx = meshp.gridFx[:, 0] <= src_b[0]\n                sgh_indz = (meshp.gridFx[:, 2] > meshp.h[2].min()) & (\n                    meshp.gridFx[:, 2] < 2 * meshp.h[2].min()\n                )\n                sgh_ind = sgh_indx & sgh_indz\n                dg_x[sgh_ind] = -1.0\n\n                # return electrode\n                sgv_indx = (meshp.gridFz[:, 0] > src_b[0] * 0.9) & (\n                    meshp.gridFz[:, 0] < src_b[0] * 1.1\n                )\n                sgv_indz = (meshp.gridFz[:, 2] >= -meshp.h[2].min()) & (\n                    meshp.gridFz[:, 2] < 2 * meshp.h[2].min()\n                )\n                sgv_ind = sgv_indx & sgv_indz\n                dg_z[sgv_ind] = 1.0\n\n                # assemble the source (downhole grounded primary)\n                dg = np.hstack([dg_x, dg_y, dg_z])\n                dg_p = [\n                    FDEM.Src.RawVec_e([], _, dg / meshp.face_areas) for _ in self.freqs\n                ]\n\n                # if plotIt:\n                #     # Plot the source to make sure the path is infact\n                #     # connected\n\n                #     fig, ax = plt.subplots(1, 1, figsize=(6, 4))\n                #     meshp.plot_grid(ax=ax)\n                #     ax.plot(meshp.gridFz[dgv_ind, 0], meshp.gridFz[dgv_ind, 2], 'rd')\n                #     ax.plot(meshp.gridFx[dgh_ind2, 0], meshp.gridFx[dgh_ind2, 2], 'rd')\n                #     ax.plot(meshp.gridFz[sgv_ind, 0], meshp.gridFz[sgv_ind, 2], 'rd')\n                #     ax.plot(meshp.gridFx[sgh_ind, 0], meshp.gridFx[sgh_ind, 2], 'rd')\n\n                #     ax.set_title('downhole casing source on mesh')\n\n                #     ax.set_xlim([0, 1.1e4])\n                #     ax.set_ylim([-1100., 0.5])\n\n                return dg_p\n\n            source_list = setupPrimarySource()  # create primary source\n            self._primarySurvey = FDEM.Survey(source_list)  # primary survey\n            print(\"... done building primary survey\")\n        return self._primarySurvey\n\n    def solvePrimary(self, primaryProblem, m=None, saveFields=False):\n        if m is None:\n            m = self.mtrue\n        print(\"solving primary ...\")\n        t0 = time.time()\n        primfields = primaryProblem.fields(m)\n        t1 = time.time()\n        print(\"Done solving primary fields, time {} \".format(t1 - t0))\n\n        return primfields\n\n    def plotPrimaryMesh(self):\n        fig, ax = plt.subplots(1, 1, figsize=(6, 4))\n        self.meshp.plot_grid(ax=ax)\n        plt.title(\"Cyl Mesh\")\n        return ax\n\n    def plotPrimaryProperties(self):\n        fig, ax = plt.subplots(1, 2, figsize=(10, 4))\n        f = self.meshp.plot_image(\n            self.muModel / mu_0,\n            ax=ax[0],\n            pcolor_opts={\"cmap\": plt.get_cmap(\"viridis\")},\n            grid=False,\n        )\n        plt.colorbar(f[0], ax=ax[0])\n        ax[0].set_xlim([0, 1.0])\n        ax[0].set_ylim([-1.5e3, 500])\n        ax[0].set_title(\"mu_r\")\n\n        f = self.meshp.plot_image(\n            np.log10(self.primaryMapping * self.mtrue),\n            ax=ax[1],\n            pcolor_opts={\"cmap\": plt.get_cmap(\"viridis\")},\n            grid=False,\n        )\n        plt.colorbar(f[0], ax=ax[1])\n        ax[1].set_xlim([0, 1.0])\n        ax[1].set_ylim([-1.5e3, 500])\n        ax[1].set_title(\"log10 sigma\")\n\n        plt.tight_layout()\n        return ax\n\n    # ----------------------------------------------------------------- #\n    # -------------- SECONDARY PROBLEM SETUP -------------------------- #\n    # ----------------------------------------------------------------- #\n\n    # -------------- MESH  -------------------------------------------- #\n\n    @property\n    def meshs(self):\n        if getattr(self, \"_meshs\", None) is None:\n            csx, ncx, npadx = 50, 21, 12\n            csy, ncy, npady = 50, 21, 12\n            csz, ncz, npadz = 25, 40, 14\n            pf = 1.5\n\n            hx = utils.unpack_widths([(csx, npadx, -pf), (csx, ncx), (csx, npadx, pf)])\n            hy = utils.unpack_widths([(csy, npady, -pf), (csy, ncy), (csy, npady, pf)])\n            hz = utils.unpack_widths([(csz, npadz, -pf), (csz, ncz), (csz, npadz, pf)])\n\n            x0 = np.r_[-hx.sum() / 2.0, -hy.sum() / 2.0, -hz[: npadz + ncz].sum()]\n            self._meshs = discretize.TensorMesh([hx, hy, hz], x0=x0)\n\n            print(\"Secondary Mesh ... \")\n            print(\n                \" xmin, xmax, zmin, zmax: \",\n                self._meshs.cell_centers_x.min(),\n                self._meshs.cell_centers_x.max(),\n                self._meshs.cell_centers_y.min(),\n                self._meshs.cell_centers_y.max(),\n                self._meshs.cell_centers_z.min(),\n                self._meshs.cell_centers_z.max(),\n            )\n            print(\" nC, vnC\", self._meshs.nC, self._meshs.vnC)\n\n        return self._meshs\n\n    @property\n    def indActive(self):\n        return self.meshs.gridCC[:, 2] <= 0.0  # air cells\n\n    @property\n    def injActMap(self):\n        return maps.InjectActiveCells(self.meshs, self.indActive, np.log(self.sigmaair))\n\n    @property\n    def expMap(self):\n        return maps.ExpMap(self.meshs)\n\n    @property\n    def mapping(self):\n        # secondary mapping\n        # here, we construct the parametric mapping to take the parameters\n        # describing the block in a layered space and map it to a conductivity\n        # model on our mesh\n        if getattr(self, \"_mapping\", None) is None:\n            print(\"building secondary mapping\")\n            paramMap = maps.ParametricBlockInLayer(self.meshs, indActive=self.indActive)\n            self._mapping = (\n                self.expMap\n                * self.injActMap  # log sigma --> sigma\n                * paramMap  # inject air cells  # block in a layered space (subsurface)\n            )\n            print(\"... done building secondary mapping\")\n\n        return self._mapping\n\n    @property\n    def primaryMap2meshs(self):\n        if getattr(self, \"_primaryMap2mesh\", None) is None:\n            # map the primary model to the secondary mesh (layer without the\n            # block)\n            print(\"Building primaryMap2meshs\")\n            paramMapPrimaryMeshs = maps.ParametricLayer(\n                self.meshs, indActive=self.indActive\n            )\n\n            self._primaryMap2mesh = (\n                self.expMap\n                * self.injActMap  # log sigma --> sigma\n                * paramMapPrimaryMeshs  # include air cells\n                * self.projectionMapPrimary  # parametrized layer  # grab correct indices\n            )\n            print(\"... done building primaryMap2meshs\")\n        return self._primaryMap2mesh\n\n    # -------------- PROBLEM and SURVEY ---------------------------- #\n    def setupSecondaryProblem(self, mapping=None):\n        print(\"Setting up Secondary Problem\")\n        if mapping is None:\n            mapping = [(\"sigma\", maps.IdentityMap(self.meshs))]\n        sec_problem = FDEM.Simulation3DElectricField(self.meshs, sigmaMap=mapping)\n        sec_problem.solver = Solver\n        print(\"... done setting up secondary problem\")\n        return sec_problem\n\n    def setupSecondarySurvey(self, primaryProblem, primarySurvey, map2meshSecondary):\n        print(\"Setting up Secondary Survey\")\n\n        nx = 41\n        ny = nx\n        rx_x, rx_y = 2 * [np.linspace(-2050, 2050, nx)]\n        self.rxlocs = utils.ndgrid([rx_x, rx_y, np.r_[-1]])\n        self.rx_x = self.rxlocs[:, 0].reshape(nx, ny, order=\"F\")\n        self.rx_y = self.rxlocs[:, 1].reshape(nx, ny, order=\"F\")\n\n        rx_ex = FDEM.Rx.PointElectricField(\n            self.rxlocs, orientation=\"x\", component=\"real\"\n        )\n        rx_ey = FDEM.Rx.PointElectricField(\n            self.rxlocs, orientation=\"y\", component=\"real\"\n        )\n\n        RxList = [rx_ex, rx_ey]\n\n        sec_src = [\n            FDEM.Src.PrimSecMappedSigma(\n                RxList,\n                freq,\n                primaryProblem,\n                primarySurvey,\n                map2meshSecondary=map2meshSecondary,\n            )\n            for freq in self.freqs\n        ]\n        print(\"... done secondary survey\")\n\n        return FDEM.Survey(sec_src)\n\n    # -------------- SOLVE ---------------------------- #\n    def solveSecondary(self, sec_problem, sec_survey, m, plotIt=False):\n        sec_problem.survey = sec_survey\n\n        print(\"Solving Secondary\")\n        t0 = time.time()\n        fields = sec_problem.fields(m)\n        dpred = sec_problem.dpred(m, f=fields)\n        t1 = time.time()\n        print(f\" ...done.   secondary time {t1 - t0}\")\n\n        return fields, dpred\n\n    # ----------------------------------------------------------------- #\n    # ------------ PLOTTING ------------------------------------------- #\n    # ----------------------------------------------------------------- #\n\n    def plotPrimaryFields(self, primaryFields, saveFig=False):\n        # Interpolate onto a cartesian mesh with uniform cell sizes (better for\n        # streamplots)\n        cs = 5.0\n        xmax = 1000.0\n        zmax = 1200.0\n        csx, ncx = cs, np.ceil(xmax / cs)\n        csz, ncz = cs, np.ceil(zmax / cs)\n\n        # define the tensor mesh\n        meshcart = discretize.TensorMesh(\n            [[(csx, ncx)], [(csx, 1)], [(csz, ncz)]], [0, -csx / 2.0, -zmax]\n        )\n\n        projF = self.meshp.get_interpolation_matrix_cartesian_mesh(meshcart, \"F\")\n\n        jcart = projF * primaryFields[:, \"j\"]\n\n        fig, ax = plt.subplots(1, 1, figsize=(6, 7.75))\n        f = meshcart.plot_slice(\n            jcart.real,\n            normal=\"Y\",\n            v_type=\"F\",\n            view=\"vec\",\n            pcolor_opts={\"norm\": LogNorm(), \"cmap\": plt.get_cmap(\"viridis\")},\n            stream_opts={\"color\": \"k\", \"arrowsize\": 2},\n            ax=ax,\n        )\n        plt.colorbar(f[0], label=\"real current density (A/m$^2$)\")\n\n        ax.set_adjustable(\"box\")\n        ax.axis(\"equal\")\n        ax.set_ylim([-1200.0, 0.0])\n        ax.set_xlim([0.0, 750.0])\n        ax.set_title(\"Primary Current Density\")\n        ax.set_xlabel(\"radius (m)\", fontsize=fontsize)\n        ax.set_ylabel(\"z (m)\", fontsize=fontsize)\n\n        if saveFig is True:\n            fig.savefig(\"primaryCurrents\", dpi=300, bbox_inches=\"tight\")\n\n        return ax\n\n    def plotSecondarySource(self, primaryFields, saveFig=False):\n        # get source term\n        secondaryProblem = self.setupSecondaryProblem(mapping=self.mapping)\n        secondaryProblem.solver = Solver\n        self.primaryProblem.solver = Solver\n        secondaryProblem.model = self.mtrue\n        secondarySurvey = self.setupSecondarySurvey(\n            self.primaryProblem, self.primarySurvey, self.primaryMap2meshs\n        )\n        src = secondarySurvey.source_list[0]\n        s_e = src.s_e(secondaryProblem, f=primaryFields)\n\n        # Mesh to interpolate onto for stream plots\n        cs = 5.0\n        csz = 0.5\n        xmin, xmax = -600.0, 600.0\n        ymin, ymax = -600.0, 600.0\n        zmin, zmax = -950.0 - csz / 2.0, -950.0 + csz / 2.0\n\n        ncx = np.ceil((xmax - xmin) / cs)\n        ncy = np.ceil((ymax - ymin) / cs)\n        ncz = np.ceil((zmax - zmin) / cs)\n\n        meshs_plt = discretize.TensorMesh(\n            [[(cs, ncx)], [(cs, ncy)], [(cs, ncz)]],\n            [\n                xmin + (xmin + xmax) / 2.0,\n                ymin + (ymin + ymax) / 2.0,\n                zmin + (zmin + zmax) / 2.0,\n            ],\n        )\n\n        # Construct interpolation matrices\n        Px = self.meshs.get_interpolation_matrix(meshs_plt.gridEx, location_type=\"Ex\")\n        Py = self.meshs.get_interpolation_matrix(meshs_plt.gridEy, location_type=\"Ey\")\n        Pz = self.meshs.get_interpolation_matrix(meshs_plt.gridEz, location_type=\"Ez\")\n        P = sp.vstack([Px, Py, Pz])\n\n        # for regions outside of the anomalous block, the source current\n        # density is identically zero. For plotting, we do not want to\n        # interpolate into this region, so we build up masked arrays.\n        maskme_ex = (\n            (self.meshs.gridEx[:, 0] <= self.block_x[0])\n            | (self.meshs.gridEx[:, 0] >= self.block_x[1])\n            | (self.meshs.gridEx[:, 1] <= self.block_y[0])\n            | (self.meshs.gridEx[:, 1] >= self.block_y[1])\n        )\n\n        maskme_ey = (\n            (self.meshs.gridEy[:, 0] <= self.block_x[0])\n            | (self.meshs.gridEy[:, 0] >= self.block_x[1])\n            | (self.meshs.gridEy[:, 1] <= self.block_y[0])\n            | (self.meshs.gridEy[:, 1] >= self.block_y[1])\n        )\n\n        maskme_ez = (\n            (self.meshs.gridEz[:, 0] <= self.block_x[0])\n            | (self.meshs.gridEz[:, 0] >= self.block_x[1])\n            | (self.meshs.gridEz[:, 1] <= self.block_y[0])\n            | (self.meshs.gridEz[:, 1] >= self.block_y[1])\n        )\n\n        maskme_e = np.hstack([maskme_ex, maskme_ey, maskme_ez])\n\n        # interpolate down a layer\n        s_e_interp = s_e.real.copy()\n        s_e_interp[maskme_e] = np.nan\n        s_e_plt = P * s_e_interp\n\n        # keep masked array for stream plots\n        s_e_stream_cc = meshs_plt.aveE2CCV * s_e_plt\n\n        # re-assign zero for amplitude of the real current density\n        s_e_abs_cc = s_e_stream_cc.reshape(meshs_plt.nC, 3, order=\"F\")\n        s_e_abs_cc = np.sqrt((s_e_abs_cc**2.0).sum(axis=1))\n        s_e_abs_cc[np.isnan(s_e_abs_cc)] = 0.0\n        s_e_stream_cc = np.ma.masked_where(np.isnan(s_e_stream_cc), s_e_stream_cc)\n\n        # plot\n        fig, ax = plt.subplots(1, 1, figsize=(7.5, 6))\n\n        # f = meshs_plt.plot_slice(\n        #     np.ma.masked_where(maskme_e, s_e_plt.real),\n        #     normal='Z',\n        #     v_type='CCv',\n        #     view='abs',\n        #     pcolor_opts={'cmap':plt.get_cmap('viridis')}, ax=ax\n        # )\n\n        f = ax.pcolormesh(\n            meshs_plt.cell_centers_x,\n            meshs_plt.cell_centers_y,\n            (s_e_abs_cc).reshape(meshs_plt.vnC[:2], order=\"F\").T,\n            cmap=plt.get_cmap(\"viridis\"),\n        )\n\n        ax.streamplot(\n            meshs_plt.cell_centers_x,\n            meshs_plt.cell_centers_y,\n            s_e_stream_cc[: meshs_plt.nC].reshape(meshs_plt.vnC[:2]),\n            s_e_stream_cc[meshs_plt.nC : meshs_plt.nC * 2].reshape(meshs_plt.vnC[:2]),\n            density=1.5,\n            color=\"k\",\n            arrowsize=2,\n        )\n\n        ax.set_xlabel(\"x (m)\", fontsize=fontsize)\n        ax.set_ylabel(\"y (m)\", fontsize=fontsize)\n\n        cb = plt.colorbar(f, label=\"real current density (A/m$^2$)\")\n        cb.formatter.set_powerlimits((0, 0))\n        cb.update_ticks()\n\n        ax.set_adjustable(\"box\")\n        ax.axis(\"equal\")\n        ax.axis([-600, 600, -600, 600])\n        ax.set_title(\"(a) -950m Depth Slice\", fontsize=fontsize)\n        # interact(plotMe, ind=[0, meshs_plt.vnC[2]-1])\n\n        if saveFig is True:\n            fig.savefig(\"secondarySource\", dpi=300)\n\n        return ax\n\n    def plotData(self, data_block, data_back, saveFig=False):\n        XLIM = np.r_[-1500, 1500]\n        YLIM = np.r_[-1500, 1500]\n\n        sec_survey = self.setupSecondarySurvey(\n            self.primaryProblem, self.primarySurvey, self.primaryMap2meshs\n        )\n        src = sec_survey.source_list[0]\n        rx0 = src.receiver_list[0]\n\n        nx = int(np.sqrt(len(rx0.locations)))\n        ny = nx\n\n        def plotDataFun(\n            ax,\n            plotme,\n            num=50,\n            plotBlock=True,\n            xlim=XLIM,\n            ylim=YLIM,\n            clim=None,\n            clabel=\"Electric Field (V/m)\",\n            xlabel=\"x (m)\",\n            ylabel=\"y (m)\",\n            title=None,\n        ):\n            if clim is None:\n                clim = np.absolute(plotme).max() * np.r_[-1.0, 1.0]\n            elif clim is not None:\n                clim = clim\n\n            f = ax.contourf(\n                self.rx_x,\n                self.rx_y,\n                plotme,\n                num,\n                cmap=plt.get_cmap(\"viridis\"),\n                vmin=clim[0],\n                vmax=clim[1],\n            )\n            ax.set_adjustable(\"box\")\n            ax.axis(\"equal\")\n            ax.set_xlim(xlim)\n            ax.set_ylim(ylim)\n            cb = plt.colorbar(f, ax=ax, label=clabel)\n            cb.formatter.set_powerlimits((0, 0))\n            cb.update_ticks()\n\n            if xlabel is not None:\n                ax.set_xlabel(xlabel)\n            if ylabel is not None:\n                ax.set_ylabel(ylabel)\n            if title is not None:\n                ax.set_title(title)\n\n            if plotBlock:\n                ax.plot(\n                    np.r_[\n                        self.block_x[0],\n                        self.block_x[0],\n                        self.block_x[1],\n                        self.block_x[1],\n                        self.block_x[0],\n                    ],\n                    np.r_[\n                        self.block_y[0],\n                        self.block_y[1],\n                        self.block_y[1],\n                        self.block_y[0],\n                        self.block_y[0],\n                    ],\n                    color=\"w\",\n                    linestyle=\"-\",\n                )\n            return ax\n\n        ncontours = 50\n\n        fig, ax = plt.subplots(2, 2, figsize=(12, 10))\n        ax = utils.mkvc(ax)\n\n        plotx0 = (data_block[: rx0.nD]).reshape(nx, ny, order=\"F\")\n        ploty0 = (data_block[rx0.nD :]).reshape(nx, ny, order=\"F\")\n\n        plotx1 = (data_block[: rx0.nD] - data_back[: rx0.nD]).reshape(nx, ny, order=\"F\")\n        ploty1 = (data_block[rx0.nD :] - data_back[rx0.nD :]).reshape(nx, ny, order=\"F\")\n\n        # Plotting\n        ax[0] = plotDataFun(ax[0], plotx0, num=ncontours, title=\"(a) Total E$_x$\")\n        ax[1] = plotDataFun(ax[1], plotx1, num=ncontours, title=\"(c) Secondary E$_x$\")\n        ax[2] = plotDataFun(ax[2], ploty0, num=ncontours, title=\"(b) Total E$_y$\")\n        ax[3] = plotDataFun(ax[3], ploty1, num=ncontours, title=\"(d) Secondary E$_y$\")\n        plt.tight_layout()\n\n        if saveFig is True:\n            fig.savefig(\"casingDpred\", dpi=300)\n\n    def plotSensitivities(self, J, saveFig=False):\n        def plotJ(\n            ax,\n            Jv,\n            title,\n            plotGrid=False,\n            xlabel=\"x (m)\",\n            ylabel=\"y (m)\",\n            xlim=None,\n            ylim=None,\n            clim=None,\n            climCenter=True,\n            plotBlock=False,\n            num=30,\n            norm=None,\n            cblabel=\"\",\n        ):\n            ax.axis(\"equal\")\n            vlim = np.absolute(Jv).max() * np.r_[-1.0, 1.0]\n\n            if norm is None:\n                f = ax.contourf(\n                    self.rx_x,\n                    self.rx_y,\n                    Jv,\n                    levels=np.linspace(vlim[0], vlim[1], num),\n                    cmap=plt.get_cmap(\"viridis\"),\n                    vmin=vlim[0],\n                    vmax=vlim[1],\n                )\n                cb = plt.colorbar(f, ax=ax, label=cblabel)\n                cb.formatter.set_powerlimits((0, 0))\n                ticks = [\n                    \"{0:1.1e}\".format(a)\n                    for a in np.linspace(0.95 * vlim[0], 0.95 * vlim[1], 5)\n                ]\n                ticks = [float(t) for t in ticks]\n                cb.set_ticks(ticks)\n                cb.update_ticks()\n\n            elif norm.lower() == \"lognorm\":\n                from matplotlib.colors import LogNorm\n\n                f = ax.contourf(\n                    self.rx_x,\n                    self.rx_y,\n                    np.absolute(Jv),\n                    num,\n                    cmap=plt.get_cmap(\"viridis\"),\n                    norm=LogNorm(),\n                )\n                cb = plt.colorbar(f, ax=ax)\n\n            ax.set_title(title)\n            ax.set_xlabel(xlabel)\n            ax.set_ylabel(ylabel)\n\n            if plotGrid:\n                self.meshs.plot_slice(\n                    np.nan * np.ones(self.meshs.nC), normal=\"Z\", grid=True, ax=ax\n                )\n\n            if xlim is not None:\n                ax.set_xlim(xlim)\n\n            if ylim is not None:\n                ax.set_ylim(ylim)\n\n            if plotBlock is True:\n                ax.plot(\n                    np.r_[\n                        self.block_x[0],\n                        self.block_x[0],\n                        self.block_x[1],\n                        self.block_x[1],\n                        self.block_x[0],\n                    ],\n                    np.r_[\n                        self.block_y[0],\n                        self.block_y[1],\n                        self.block_y[1],\n                        self.block_y[0],\n                        self.block_y[0],\n                    ],\n                    color=\"w\",\n                    linestyle=\"-\",\n                )\n\n            return ax\n\n        # Plot Conductivity contribution\n        plotGrid = False\n        plotBlock = True\n        ncontours = 30\n\n        xlim = np.r_[-1500, 1500]\n        ylim = np.r_[-1500, 1500]\n\n        nx, ny = self.rx_x.shape\n        nrx = len(self.rxlocs)\n\n        J_back_ex = J[0, :nrx].reshape(nx, ny, order=\"F\")\n        J_back_ey = J[0, nrx:].reshape(nx, ny, order=\"F\")\n        J_layer_ex = J[1, :nrx].reshape(nx, ny, order=\"F\")\n        J_layer_ey = J[1, nrx:].reshape(nx, ny, order=\"F\")\n        J_block_ex = J[2, :nrx].reshape(nx, ny, order=\"F\")\n        J_block_ey = J[2, nrx:].reshape(nx, ny, order=\"F\")\n\n        clabelSigs = r\"Sensitivity (V/m / log($\\sigma$))\"\n\n        fig, ax = plt.subplots(3, 2, figsize=(12, 15))\n        ax[0][0] = plotJ(\n            ax[0][0],\n            J_back_ex,\n            r\"(a) Sensitivity of $E_x$ wrt log($\\sigma_{back}$)\",\n            plotGrid=plotGrid,\n            xlim=xlim,\n            ylim=ylim,\n            plotBlock=plotBlock,\n            num=ncontours,\n            cblabel=clabelSigs,\n        )\n\n        ax[0][1] = plotJ(\n            ax[0][1],\n            J_back_ey,\n            r\"(b) Sensitivity of $E_y$ wrt log($\\sigma_{back}$)\",\n            plotGrid=plotGrid,\n            xlim=xlim,\n            ylim=ylim,\n            plotBlock=plotBlock,\n            num=ncontours,\n            cblabel=clabelSigs,\n        )\n\n        ax[1][0] = plotJ(\n            ax[1][0],\n            J_layer_ex,\n            r\"(c) Sensitivity of $E_x$ wrt log($\\sigma_{layer}$)\",\n            plotGrid=plotGrid,\n            xlim=xlim,\n            ylim=ylim,\n            plotBlock=plotBlock,\n            num=ncontours,\n            cblabel=clabelSigs,\n        )\n\n        ax[1][1] = plotJ(\n            ax[1][1],\n            J_layer_ey,\n            r\"(d) Sensitivity of $E_y$ wrt log($\\sigma_{layer}$)\",\n            plotGrid=plotGrid,\n            xlim=xlim,\n            ylim=ylim,\n            plotBlock=plotBlock,\n            num=ncontours,\n            cblabel=clabelSigs,\n        )\n\n        climsigblock = np.r_[-6e-8, 6e-8]\n        ax[2][0] = plotJ(\n            ax[2][0],\n            J_block_ex,\n            r\"(e) Sensitivity of $E_x$ wrt log($\\sigma_{block}$)\",\n            plotGrid=plotGrid,\n            xlim=xlim,\n            ylim=ylim,\n            clim=climsigblock,\n            plotBlock=plotBlock,\n            num=ncontours,\n            cblabel=clabelSigs,\n        )\n\n        ax[2][1] = plotJ(\n            ax[2][1],\n            J_block_ey,\n            r\"(f) Sensitivity of $E_y$ wrt log($\\sigma_{block}$)\",\n            plotGrid=plotGrid,\n            xlim=xlim,\n            ylim=ylim,\n            clim=climsigblock,\n            plotBlock=plotBlock,\n            num=ncontours,\n            cblabel=clabelSigs,\n        )\n\n        plt.tight_layout()\n\n        if saveFig is True:\n            fig.savefig(\"J_sigmas\", dpi=300)\n\n        # Plot layer contribution\n        fig, ax = plt.subplots(2, 2, figsize=(12, 10))\n        # ax = utils.mkvc(ax)\n\n        xlim = np.r_[-1500.0, 1500.0]\n        ylim = np.r_[-1500.0, 1500.0]\n\n        J_z0_ex, J_z0_ey = (\n            J[3, :nrx].reshape(nx, ny, order=\"F\"),\n            J[3, nrx:].reshape(nx, ny, order=\"F\"),\n        )\n        J_hz_ex, J_hz_ey = (\n            J[4, :nrx].reshape(nx, ny, order=\"F\"),\n            J[4, nrx:].reshape(nx, ny, order=\"F\"),\n        )\n\n        ax[0][0] = plotJ(\n            ax[0][0],\n            J_z0_ex,\n            \"(g) Sensitivity of $E_x$ wrt layer $z_0$\",\n            plotGrid=plotGrid,\n            xlim=xlim,\n            ylim=ylim,\n            plotBlock=plotBlock,\n            num=ncontours,\n            cblabel=\"Sensitivity (V/m / m)\",\n        )\n\n        ax[0][1] = plotJ(\n            ax[0][1],\n            J_z0_ey,\n            \"(h) Sensitivity of $E_y$ wrt layer $z_0$\",\n            plotGrid=plotGrid,\n            xlim=xlim,\n            ylim=ylim,\n            plotBlock=plotBlock,\n            num=ncontours,\n            cblabel=\"Sensitivity (V/m / m)\",\n        )\n\n        ax[1][0] = plotJ(\n            ax[1][0],\n            J_hz_ex,\n            \"(i) Sensitivity of $E_x$ wrt layer $h$\",\n            plotGrid=plotGrid,\n            xlim=xlim,\n            ylim=ylim,\n            plotBlock=plotBlock,\n            num=ncontours,\n            cblabel=\"Sensitivity (V/m / m)\",\n        )\n\n        ax[1][1] = plotJ(\n            ax[1][1],\n            J_hz_ey,\n            \"(j) Sensitivity of $E_y$ wrt layer $h$\",\n            plotGrid=plotGrid,\n            xlim=xlim,\n            ylim=ylim,\n            plotBlock=plotBlock,\n            num=ncontours,\n            cblabel=\"Sensitivity (V/m / m)\",\n        )\n\n        plt.tight_layout()\n\n        if saveFig is True:\n            fig.savefig(\"J_layer\", dpi=300)\n\n        # Block Geometry\n        fig, ax = plt.subplots(4, 2, figsize=(12, 20))\n\n        xlim = np.r_[-1500.0, 1500.0]\n        ylim = np.r_[-1500.0, 1500.0]\n\n        J_x0_ex = J[5, :nrx].reshape(nx, ny, order=\"F\")\n        J_x0_ey = J[5, nrx:].reshape(nx, ny, order=\"F\")\n        J_y0_ex = J[6, :nrx].reshape(nx, ny, order=\"F\")\n        J_y0_ey = J[6, nrx:].reshape(nx, ny, order=\"F\")\n        J_dx_ex = J[7, :nrx].reshape(nx, ny, order=\"F\")\n        J_dx_ey = J[7, nrx:].reshape(nx, ny, order=\"F\")\n        J_dy_ex = J[8, :nrx].reshape(nx, ny, order=\"F\")\n        J_dy_ey = J[8, nrx:].reshape(nx, ny, order=\"F\")\n\n        ax[0][0] = plotJ(\n            ax[0][0],\n            J_x0_ex,\n            \"(k) Sensitivity of $E_x$ wrt block $x_0$\",\n            plotGrid=plotGrid,\n            xlim=xlim,\n            ylim=ylim,\n            plotBlock=plotBlock,\n            num=ncontours,\n            cblabel=\"Sensitivity (V/m / m)\",\n        )\n\n        ax[0][1] = plotJ(\n            ax[0][1],\n            J_x0_ey,\n            \"(l) Sensitivity of $E_y$ wrt block $x_0$\",\n            plotGrid=plotGrid,\n            xlim=xlim,\n            ylim=ylim,\n            plotBlock=plotBlock,\n            num=ncontours,\n            cblabel=\"Sensitivity (V/m / m)\",\n        )\n\n        ax[1][0] = plotJ(\n            ax[1][0],\n            J_y0_ex,\n            \"(m) Sensitivity of $E_x$ wrt block $y_0$\",\n            plotGrid=plotGrid,\n            xlim=xlim,\n            ylim=ylim,\n            plotBlock=plotBlock,\n            num=ncontours,\n            cblabel=\"Sensitivity (V/m / m)\",\n        )\n\n        ax[1][1] = plotJ(\n            ax[1][1],\n            J_y0_ey,\n            \"(n) Sensitivity of $E_y$ wrt block $y_0$\",\n            plotGrid=plotGrid,\n            xlim=xlim,\n            ylim=ylim,\n            plotBlock=plotBlock,\n            num=ncontours,\n            cblabel=\"Sensitivity (V/m / m)\",\n        )\n\n        ax[2][0] = plotJ(\n            ax[2][0],\n            J_dx_ex,\n            \"(o) Sensitivity of $E_x$ wrt block $d_x$\",\n            plotGrid=plotGrid,\n            xlim=xlim,\n            ylim=ylim,\n            plotBlock=plotBlock,\n            num=ncontours,\n            cblabel=\"Sensitivity (V/m / m)\",\n        )\n\n        ax[2][1] = plotJ(\n            ax[2][1],\n            J_dx_ey,\n            \"(p) Sensitivity of $E_y$ wrt block $d_x$\",\n            plotGrid=plotGrid,\n            xlim=xlim,\n            ylim=ylim,\n            plotBlock=plotBlock,\n            num=ncontours,\n            cblabel=\"Sensitivity (V/m / m)\",\n        )\n\n        ax[3][0] = plotJ(\n            ax[3][0],\n            J_dy_ex,\n            \"(q) Sensitivity of $E_x$ wrt block $d_y$\",\n            plotGrid=plotGrid,\n            xlim=xlim,\n            ylim=ylim,\n            plotBlock=plotBlock,\n            num=ncontours,\n            cblabel=\"Sensitivity (V/m / m)\",\n        )\n\n        ax[3][1] = plotJ(\n            ax[3][1],\n            J_dy_ey,\n            \"(r) Sensitivity of $E_y$ wrt block $d_y$\",\n            plotGrid=plotGrid,\n            xlim=xlim,\n            ylim=ylim,\n            plotBlock=plotBlock,\n            num=ncontours,\n            cblabel=\"Sensitivity (V/m / m)\",\n        )\n\n        plt.tight_layout()\n\n        if saveFig is True:\n            fig.savefig(\"J_block\", dpi=300)\n\n    # ---------------------------------------------------------------------- #\n    # ---------------- Run the example ------------------------------------- #\n    # ---------------------------------------------------------------------- #\n\n    def run(\n        self, plotIt=False, runTests=False, verbose=True, saveFields=True, saveFig=False\n    ):\n        self.verbose = verbose\n\n        if plotIt is True:  # Plot the Primary Model\n            # self.plotPrimaryMesh() # plot the mesh\n            self.plotPrimaryProperties()  # plot mu, sigma\n\n        # Primary Simulation\n        self.primaryProblem.survey = self.primarySurvey\n        primfields = self.solvePrimary(self.primaryProblem, m=self.mtrue)\n\n        if saveFields is True:\n            np.save(\"primaryfields_\" + self.NAME, primfields[:, :])\n            print(\"   saved %s\" % \"primaryfields_\" + self.NAME)\n\n        mback = self.mtrue.copy()\n        mback[2] = np.log(self.sigmalayer)\n\n        # Secondary Problem and Survey\n        sec_problem = self.setupSecondaryProblem(mapping=self.mapping)\n\n        sec_survey = self.setupSecondarySurvey(\n            self.primaryProblem, self.primarySurvey, self.primaryMap2meshs\n        )\n        sec_problem.survey = sec_survey\n\n        # layered earth only (background)\n        background_problem = self.setupSecondaryProblem(mapping=self.primaryMap2meshs)\n        background_survey = self.setupSecondarySurvey(\n            self.primaryProblem, self.primarySurvey, self.primaryMap2meshs\n        )\n        background_problem.survey = background_survey\n\n        # -------------- Test the sensitivity ----------------------------- #\n        if runTests:\n            # Test Block Model\n            def fun(x):\n                return [sec_problem.dpred(x), lambda x: sec_problem.Jvec(self.mtrue, x)]\n\n            tests.check_derivative(fun, self.mtrue, num=2, plotIt=False)\n\n        # -------------- Calculate Fields --------------------------------- #\n        # Background\n        t0 = time.time()\n        print(\"solving background ... \")\n        fieldsback, dpredback = self.solveSecondary(\n            background_problem, background_survey, self.mtrue\n        )\n        t1 = time.time()\n        print(\"... done.   dpred_back {}\".format(t1 - t0))\n\n        if saveFields:\n            np.save(\"dpred_\" + self.NAME + \"_back\", dpredback)\n            np.save(\"fields_\" + self.NAME + \"_back\", fieldsback[:, :])\n\n            print(\"   saved {}\".format(self.NAME + \"_back\"))\n\n        # with Block\n        t0 = time.time()\n        print(\"solving with block ... \")\n        fields, dpred = self.solveSecondary(sec_problem, sec_survey, self.mtrue)\n        print(\"... done.   dpred {}\".format(t1 - t0))\n        if saveFields:\n            np.save(\"dpred_\" + self.NAME, dpred)\n            np.save(\"fields_\" + self.NAME, fields[:, :])\n            print(\"   saved {}\".format(self.NAME))\n\n        t1 = time.time()\n\n        # -------------- Calculate J --------------------------------- #\n        # Calculate J with block\n        print(\"starting J with block\")\n        t0 = time.time()\n\n        J = []\n        for i in range(len(self.mtrue)):\n            ei = np.zeros_like(self.mtrue)\n            ei[i] = 1.0\n            J.append(sec_problem.Jvec(self.mtrue, ei, f=fields))\n\n        J = np.vstack(J)\n\n        t1 = time.time()\n        print(\"   J {}\".format(t1 - t0))\n\n        if saveFields is True:\n            np.save(\"J_\" + self.NAME, J)\n            print(\"   saved {}\".format(\"J_\" + self.NAME))\n\n        return {\n            \"primfields\": primfields,  # primary fields\n            \"fieldsback\": fieldsback,  # fields without block\n            \"dpredback\": dpredback,  # predicted data without block\n            \"fields\": fields,  # fields with block\n            \"dpred\": dpred,  # predicted data with block\n            \"J\": J,  # sensitivity\n        }\n\n\nclass PrimSecCasingStoredResults(PrimSecCasingExample):\n    url = \"https://storage.googleapis.com/simpeg/papers/Heagyetal2016/\"\n\n    # cloudfiles = [\n    #     'primaryfields_PrimSec_5e6Casing_50Mu_05Hz_LargeCondBody.npy',\n    #     'dpred_PrimSec_5e6Casing_50Mu_05Hz_LargeCondBody_back.npy',\n    #     'dpred_PrimSec_5e6Casing_50Mu_05Hz_LargeCondBody.npy',\n    #     'J_PrimSec_5e6Casing_50Mu_05Hz_LargeCondBody.npy',\n    #              ]\n\n    cloudfile = \"Heagyetal2016Casing.hdf5\"\n\n    entry_names = [\n        \"primaryfields_h_PrimSec_5e6Casing_50Mu_05Hz_LargeCondBody\",\n        \"dpred_PrimSec_5e6Casing_50Mu_05Hz_LargeCondBody_back\",\n        \"dpred_PrimSec_5e6Casing_50Mu_05Hz_LargeCondBody\",\n        \"J_PrimSec_5e6Casing_50Mu_05Hz_LargeCondBody\",\n    ]\n\n    def removeStoredResults(self):\n        import shutil\n\n        print(\"Removing {}\".format(self.filepath))\n        shutil.rmtree(self.filepath)\n\n    def run(self, plotIt=False, runTests=False, saveFig=False):\n        filepath = download(\n            self.url + self.cloudfile, folder=\"~/Downloads/simpegtemp\", overwrite=True\n        )\n        self.filepath = os.path.sep.join(filepath.split(os.path.sep)[:-1])\n\n        # resultsFiles = ['{filepath}{slash}{file}'.format(\n        #     filepath=self.filepath, slash=os.path.sep, file=file)\n        #     for file in self.cloudfiles]\n        # results = [np.load(file, encoding='bytes') for file in resultsFiles]\n\n        h5f = h5py.File(filepath, \"r\")\n        results = [h5f[entry_name][:] for entry_name in self.entry_names]\n        results = dict(zip([\"primfields\", \"dpredback\", \"dpred\", \"J\"], results))\n\n        # Put the primary fields into a fields object\n        self.primaryProblem.model = self.mtrue  # set the current model\n        self.primaryProblem.survey = self.primarySurvey\n        primaryFields = self.primaryProblem.fieldsPair(self.primaryProblem)\n        primaryFields[self.primarySurvey.source_list[0], \"hSolution\"] = results[\n            \"primfields\"\n        ]\n\n        results[\"primfields\"] = primaryFields\n\n        return results\n\n\ndef run(plotIt=True, runTests=False, reRun=False, saveFig=False):\n    \"\"\"\n    EM Heagyetal2016 CasingFwd3DPrimSecSrc\n    ======================================\n\n    Computation of Sensitivities for the primary-secondary example shown in\n    Heagy et al 2016.\n\n    :param bool plotIt: plot results\n    :param bool runTests: run sensitivity tests? (slow...)\n    :param bool reRun: recompute results? or just download stored results\n        and plot\n    :param bool saveFig: save the figures?\n    \"\"\"\n    # recompute results?\n    if reRun is True:\n        casingExample = PrimSecCasingExample()\n\n    # or download stored results\n    elif reRun is False:\n        casingExample = PrimSecCasingStoredResults()\n\n    dataDict = casingExample.run(runTests=runTests)\n\n    # plot some things\n    if plotIt is True or saveFig is True:\n        casingExample.plotPrimaryFields(dataDict[\"primfields\"], saveFig=saveFig)\n        casingExample.plotSecondarySource(dataDict[\"primfields\"], saveFig=saveFig)\n        casingExample.plotData(\n            dataDict[\"dpred\"], dataDict[\"dpredback\"], saveFig=saveFig\n        )\n        casingExample.plotSensitivities(dataDict[\"J\"], saveFig=saveFig)\n\n    if plotIt is True:\n        plt.show()\n\n    # remove the downloaded results\n    if reRun is False:\n        try:\n            casingExample.removeStoredResults()\n        except PermissionError:\n            pass\n\n\nif __name__ == \"__main__\":\n    run(plotIt=True, runTests=False, reRun=False, saveFig=False)"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.8.19"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}